@@ -27,6 +27,8 @@ using namespace std;
 #include <future>
 
 #include <boost/format.hpp>
+#include <osmium/memory/buffer.hpp>
+#include <osmium/builder/osm_object_builder.hpp>
 
 #include <libpq-fe.h>
 
@@ -96,43 +98,66 @@ inline const char *decode_upto( const char *src, char *dst )
   return src;
 }
 
-void pgsql_parse_tags(const char *string, taglist_t &tags)
+template <typename T>
+void pgsql_parse_tags(const char *string, osmium::memory::Buffer &buffer, T &obuilder)
 {
-  char key[1024];
-  char val[1024];
+    if( *string++ != '{' )
+        return;
 
-  if( *string == '\0' )
-    return;
+    char key[1024];
+    char val[1024];
+    osmium::builder::TagListBuilder builder(buffer, &obuilder);
 
-  if( *string++ != '{' )
-    return;
-  while( *string != '}' )
-  {
-    string = decode_upto( string, key );
-    // String points to the comma */
-    string++;
-    string = decode_upto( string, val );
-    // String points to the comma or closing '}' */
-    tags.push_back(tag_t(key, val));
-    if( *string == ',' )
-      string++;
-  }
+    while( *string != '}' ) {
+        string = decode_upto(string, key);
+        // String points to the comma */
+        string++;
+        string = decode_upto(string, val);
+        builder.add_tag(key, val);
+        // String points to the comma or closing '}' */
+        if( *string == ',' ) {
+            string++;
+        }
+    }
 }
 
-// Parses an array of integers */
-void pgsql_parse_nodes(const char *string, idlist_t &nds)
+void pgsql_parse_members(const char *string, osmium::memory::Buffer &buffer,
+                         osmium::builder::RelationBuilder &obuilder)
 {
-  if( *string++ != '{' )
-    return;
+    if( *string++ != '{' )
+        return;
 
-  while( *string != '}' )
-  {
-    char *ptr;
-    nds.push_back(strtoosmid( string, &ptr, 10 ));
-    string = ptr;
-    if( *string == ',' )
-      string++;
-  }
+    char role[1024];
+    osmium::builder::RelationMemberListBuilder builder(buffer, &obuilder);
+
+    while( *string != '}' ) {
+        char type = string[0];
+        char *endp;
+        osmid_t id = strtoosmid(string + 1, &endp, 10);
+        // String points to the comma */
+        string = decode_upto(endp + 1, role);
+        builder.add_member(osmium::char_to_item_type(type), id, role);
+        // String points to the comma or closing '}' */
+        if( *string == ',' ) {
+            string++;
+        }
+    }
+}
+
+void pgsql_parse_nodes(const char *string, osmium::memory::Buffer &buffer,
+                         osmium::builder::WayBuilder &builder)
+{
+    if (*string++ == '{') {
+        osmium::builder::WayNodeListBuilder wnl_builder(buffer, &builder);
+        while (*string != '}') {
+            char *ptr;
+            wnl_builder.add_node_ref(strtoosmid(string, &ptr, 10));
+            string = ptr;
+            if (*string == ',') {
+                string++;
+            }
+        }
+    }
 }
 
 int pgsql_endCopy(middle_pgsql_t::table_desc *table)
@@ -146,13 +171,11 @@ int pgsql_endCopy(middle_pgsql_t::table_desc *table)
             util::exit_nicely();
         }
 
-        PGresult *res = PQgetResult(sql_conn);
-        if (PQresultStatus(res) != PGRES_COMMAND_OK) {
+        pg_result_t res(PQgetResult(sql_conn));
+        if (PQresultStatus(res.get()) != PGRES_COMMAND_OK) {
             fprintf(stderr, "COPY_END for %s failed: %s\n", table->copy, PQerrorMessage(sql_conn));
-            PQclear(res);
             util::exit_nicely();
         }
-        PQclear(res);
         table->copyMode = 0;
     }
     return 0;
@@ -160,23 +183,6 @@ int pgsql_endCopy(middle_pgsql_t::table_desc *table)
 } // anonymous namespace
 
 
-void middle_pgsql_t::buffer_store_nodes(idlist_t const &nds)
-{
-    if (nds.size() == 0) {
-        copy_buffer += "{}";
-        return;
-    }
-
-    copy_buffer += "{";
-
-    for (auto const &it : nds) {
-        copy_buffer += std::to_string(it);
-        copy_buffer += ',';
-    }
-
-    copy_buffer[copy_buffer.size() - 1] = '}';
-}
-
 void middle_pgsql_t::buffer_store_string(std::string const &in, bool escape)
 {
     for (char const c: in) {
@@ -209,25 +215,31 @@ void middle_pgsql_t::buffer_store_string(std::string const &in, bool escape)
 }
 
 // escape means we return '\N' for copy mode, otherwise we return just nullptr
-void middle_pgsql_t::buffer_store_tags(taglist_t const &tags, bool escape)
+void middle_pgsql_t::buffer_store_tags(osmium::OSMObject const &obj, bool attrs,
+                                       bool escape)
 {
     copy_buffer += "{";
 
-    bool first = true;
-    for (auto const &it : tags) {
-        if (!first) {
-            copy_buffer += ',';
-        }
+    for (auto const &it : obj.tags()) {
         copy_buffer += "\"";
-        buffer_store_string(it.key, escape);
+        buffer_store_string(it.key(), escape);
         copy_buffer += "\",\"";
-        buffer_store_string(it.value, escape);
-        copy_buffer += '"';
-
-        first = false;
+        buffer_store_string(it.value(), escape);
+        copy_buffer += "\",";
+    }
+    if (attrs) {
+        taglist_t extra;
+        extra.add_attributes(obj);
+        for (auto const &it : extra) {
+            copy_buffer += "\"";
+            copy_buffer += it.key;
+            copy_buffer += "\",\"";
+            buffer_store_string(it.value.c_str(), escape);
+            copy_buffer += "\",";
+        }
     }
 
-    copy_buffer += "}";
+    copy_buffer[copy_buffer.size() - 1] = '}';
 }
 
 void middle_pgsql_t::buffer_correct_params(char const **param, size_t size)
@@ -242,159 +254,110 @@ void middle_pgsql_t::buffer_correct_params(char const **param, size_t size)
     }
 }
 
-void middle_pgsql_t::local_nodes_set(osmid_t id, double lat, double lon,
-                                     const taglist_t &tags)
+void middle_pgsql_t::local_nodes_set(osmium::Node const &node)
 {
-    copy_buffer.reserve(tags.size() * 24 + 64);
+    copy_buffer.reserve(node.tags().byte_size() + 100);
 
     bool copy = node_table->copyMode;
     char delim = copy ? '\t' : '\0';
-    const char *paramValues[4] = { copy_buffer.c_str(), };
+    const char *paramValues[4] = {
+        copy_buffer.c_str(),
+    };
 
-    copy_buffer = std::to_string(id);
+    copy_buffer = std::to_string(node.id());
     copy_buffer += delim;
 
-#ifdef FIXED_POINT
-    ramNode n(lon, lat);
     paramValues[1] = paramValues[0] + copy_buffer.size();
-    copy_buffer += std::to_string(n.int_lat());
+    copy_buffer += std::to_string(node.location().y());
     copy_buffer += delim;
 
     paramValues[2] = paramValues[0] + copy_buffer.size();
-    copy_buffer += std::to_string(n.int_lon());
-    copy_buffer += delim;
-#else
-    paramValues[1] = paramValues[0] + copy_buffer.size();
-    copy_buffer += std::to_string(lat);
-    copy_buffer += delim;
-
-    paramValues[2] = paramValues[0] + copy_buffer.size();
-    copy_buffer += std::to_string(lon);
-    copy_buffer += delim;
-#endif
-
-    if (tags.size() == 0) {
-        paramValues[3] = nullptr;
-        copy_buffer += "\\N";
-    } else {
-        paramValues[3] = paramValues[0] + copy_buffer.size();
-        buffer_store_tags(tags, copy);
-    }
+    copy_buffer += std::to_string(node.location().x());
 
     if (copy) {
         copy_buffer += '\n';
         pgsql_CopyData(__FUNCTION__, node_table->sql_conn, copy_buffer);
     } else {
         buffer_correct_params(paramValues, 4);
-        pgsql_execPrepared(node_table->sql_conn, "insert_node", 4,
-                           (const char * const *)paramValues, PGRES_COMMAND_OK);
+        pgsql_execPrepared(node_table->sql_conn, "insert_node", 3,
+                           (const char *const *)paramValues, PGRES_COMMAND_OK);
     }
 }
 
-// This should be made more efficient by using an IN(ARRAY[]) construct */
-size_t middle_pgsql_t::local_nodes_get_list(nodelist_t &out, const idlist_t nds) const
+size_t middle_pgsql_t::local_nodes_get_list(osmium::WayNodeList *nodes) const
 {
-    assert(out.empty());
+    size_t count = 0;
+    std::string buffer("{");
 
-    char tmp[16];
-
-    char *tmp2 = static_cast<char *>(malloc(sizeof(char) * nds.size() * 16));
-    if (tmp2 == nullptr) return 0; //failed to allocate memory, return */
-
-
-    // create a list of ids in tmp2 to query the database  */
-    sprintf(tmp2, "{");
-    int countDB = 0;
-    for(idlist_t::const_iterator it = nds.begin(); it != nds.end(); ++it) {
-        // Check cache first */
-        osmNode loc;
-        if (cache->get(&loc, *it) == 0) {
-            out.push_back(loc);
-            continue;
+    // get nodes where possible from cache,
+    // at the same time build a list for querying missing nodes from DB
+    size_t pos = 0;
+    for (auto &n : *nodes) {
+        auto loc = cache->get(n.ref());
+        if (loc.valid()) {
+            n.set_location(loc);
+            ++count;
+        } else {
+            buffer += std::to_string(n.ref());
+            buffer += ',';
         }
-
-        countDB++;
-        // Mark nodes as needing to be fetched from the DB */
-        out.push_back(osmNode());
-
-        snprintf(tmp, sizeof(tmp), "%" PRIdOSMID ",", *it);
-        strncat(tmp2, tmp, sizeof(char)*(nds.size()*16 - 2));
+        ++pos;
     }
-    tmp2[strlen(tmp2) - 1] = '}'; // replace last , with } to complete list of ids*/
 
-    if (countDB == 0) {
-        free(tmp2);
-        return nds.size(); // All ids where in cache, so nothing more to do */
+    if (count == pos) {
+        return count; // all ids found in cache, nothing more to do
     }
 
+    // get any remaining nodes from the DB
+    buffer[buffer.size() - 1] = '}';
+
     pgsql_endCopy(node_table);
 
     PGconn *sql_conn = node_table->sql_conn;
 
     char const *paramValues[1];
-    paramValues[0] = tmp2;
-    PGresult *res = pgsql_execPrepared(sql_conn, "get_node_list", 1, paramValues, PGRES_TUPLES_OK);
-    int countPG = PQntuples(res);
+    paramValues[0] = buffer.c_str();
+    auto res = pgsql_execPrepared(sql_conn, "get_node_list", 1, paramValues,
+                                  PGRES_TUPLES_OK);
+    auto countPG = PQntuples(res.get());
 
-    //store the pg results in a hashmap and telling it how many we expect
-    std::unordered_map<osmid_t, osmNode> pg_nodes(countPG);
-
-    for (int i = 0; i < countPG; i++) {
-        osmid_t id = strtoosmid(PQgetvalue(res, i, 0), nullptr, 10);
-        osmNode node;
-#ifdef FIXED_POINT
-        ramNode n((int) strtol(PQgetvalue(res, i, 2), nullptr, 10),
-                  (int) strtol(PQgetvalue(res, i, 1), nullptr, 10));
-
-        node.lat = n.lat();
-        node.lon = n.lon();
-#else
-        node.lat = strtod(PQgetvalue(res, i, 1), nullptr);
-        node.lon = strtod(PQgetvalue(res, i, 2), nullptr);
-#endif
-        pg_nodes.emplace(id, node);
+    std::unordered_map<osmid_t, osmium::Location> locs;
+    for (int i = 0; i < countPG; ++i) {
+        locs.emplace(
+            strtoosmid(PQgetvalue(res.get(), i, 0), nullptr, 10),
+            osmium::Location(
+                (int)strtol(PQgetvalue(res.get(), i, 2), nullptr, 10),
+                (int)strtol(PQgetvalue(res.get(), i, 1), nullptr, 10)));
     }
 
-    PQclear(res);
-    free(tmp2);
-
-    // If some of the nodes in the way don't exist, the returning list has holes.
-    // Merge the two lists removing any holes.
-    size_t wrtidx = 0;
-    for (size_t i = 0; i < nds.size(); ++i) {
-        if (std::isnan(out[i].lat)) {
-            std::unordered_map<osmid_t, osmNode>::iterator found = pg_nodes.find(nds[i]);
-            if(found != pg_nodes.end()) {
-                out[wrtidx] = found->second;
-                ++wrtidx;
-            }
-        } else {
-            if (wrtidx < i)
-                out[wrtidx] = out[i];
-            ++wrtidx;
+    for (auto &n : *nodes) {
+        auto el = locs.find(n.ref());
+        if (el != locs.end()) {
+            n.set_location(el->second);
+            ++count;
         }
-    }
-    out.resize(wrtidx);
 
-    return wrtidx;
+    }
+
+    return count;
 }
 
-
-void middle_pgsql_t::nodes_set(osmid_t id, double lat, double lon, const taglist_t &tags) {
-    cache->set( id, lat, lon, tags );
+void middle_pgsql_t::nodes_set(osmium::Node const &node)
+{
+    cache->set(node.id(), node.location());
 
     if (out_options->flat_node_cache_enabled) {
-        persistent_cache->set(id, lat, lon);
+        persistent_cache->set(node.id(), node.location());
     } else {
-        local_nodes_set(id, lat, lon, tags);
+        local_nodes_set(node);
     }
 }
 
-size_t middle_pgsql_t::nodes_get_list(nodelist_t &out, const idlist_t nds) const
+size_t middle_pgsql_t::nodes_get_list(osmium::WayNodeList *nodes) const
 {
     return (out_options->flat_node_cache_enabled)
-             ? persistent_cache->get_list(out, nds)
-             : local_nodes_get_list(out, nds);
+        ? persistent_cache->get_list(nodes)
+        : local_nodes_get_list(nodes);
 }
 
 void middle_pgsql_t::local_nodes_delete(osmid_t osm_id)
@@ -412,7 +375,7 @@ void middle_pgsql_t::local_nodes_delete(osmid_t osm_id)
 void middle_pgsql_t::nodes_delete(osmid_t osm_id)
 {
     if (out_options->flat_node_cache_enabled) {
-        persistent_cache->set(osm_id, NAN, NAN);
+        persistent_cache->set(osm_id, osmium::Location());
     } else {
         local_nodes_delete(osm_id);
     }
@@ -435,47 +398,53 @@ void middle_pgsql_t::node_changed(osmid_t osm_id)
 
     //keep track of whatever ways and rels these nodes intersect
     //TODO: dont need to stop the copy above since we are only reading?
-    PGresult* res = pgsql_execPrepared(way_table->sql_conn, "mark_ways_by_node", 1, paramValues, PGRES_TUPLES_OK );
-    for(int i = 0; i < PQntuples(res); ++i)
-    {
+    auto res = pgsql_execPrepared(way_table->sql_conn, "mark_ways_by_node", 1,
+                                  paramValues, PGRES_TUPLES_OK);
+    for (int i = 0; i < PQntuples(res.get()); ++i) {
         char *end;
-        osmid_t marked = strtoosmid(PQgetvalue(res, i, 0), &end, 10);
+        osmid_t marked = strtoosmid(PQgetvalue(res.get(), i, 0), &end, 10);
         ways_pending_tracker->mark(marked);
     }
-    PQclear(res);
 
     //do the rels too
     res = pgsql_execPrepared(rel_table->sql_conn, "mark_rels_by_node", 1, paramValues, PGRES_TUPLES_OK );
-    for(int i = 0; i < PQntuples(res); ++i)
-    {
+    for (int i = 0; i < PQntuples(res.get()); ++i) {
         char *end;
-        osmid_t marked = strtoosmid(PQgetvalue(res, i, 0), &end, 10);
+        osmid_t marked = strtoosmid(PQgetvalue(res.get(), i, 0), &end, 10);
         rels_pending_tracker->mark(marked);
     }
-    PQclear(res);
 }
 
-void middle_pgsql_t::ways_set(osmid_t way_id, const idlist_t &nds, const taglist_t &tags)
+void middle_pgsql_t::ways_set(osmium::Way const &way)
 {
-    copy_buffer.reserve(nds.size() * 10 + tags.size() * 24 + 64);
+    copy_buffer.reserve(way.nodes().size() * 10 + way.tags().byte_size() + 100);
     bool copy = way_table->copyMode;
     char delim = copy ? '\t' : '\0';
     // Three params: id, nodes, tags */
     const char *paramValues[4] = { copy_buffer.c_str(), };
 
-    copy_buffer = std::to_string(way_id);
+    copy_buffer = std::to_string(way.id());
     copy_buffer += delim;
 
     paramValues[1] = paramValues[0] + copy_buffer.size();
-    buffer_store_nodes(nds);
+    if (way.nodes().size() == 0) {
+        copy_buffer += "{}";
+    } else {
+        copy_buffer += "{";
+        for (auto const &n : way.nodes()) {
+            copy_buffer += std::to_string(n.ref());
+            copy_buffer += ',';
+        }
+        copy_buffer[copy_buffer.size() - 1] = '}';
+    }
     copy_buffer += delim;
 
-    if (tags.size() == 0) {
+    if (way.tags().empty() && !out_options->extra_attributes) {
         paramValues[2] = nullptr;
         copy_buffer += "\\N";
     } else {
         paramValues[2] = paramValues[0] + copy_buffer.size();
-        buffer_store_tags(tags, copy);
+        buffer_store_tags(way, out_options->extra_attributes, copy);
     }
 
     if (copy) {
@@ -488,106 +457,106 @@ void middle_pgsql_t::ways_set(osmid_t way_id, const idlist_t &nds, const taglist
     }
 }
 
-bool middle_pgsql_t::ways_get(osmid_t id, taglist_t &tags, nodelist_t &nodes) const
+bool middle_pgsql_t::ways_get(osmid_t id, osmium::memory::Buffer &buffer) const
 {
     char const *paramValues[1];
     PGconn *sql_conn = way_table->sql_conn;
 
     // Make sure we're out of copy mode */
-    pgsql_endCopy( way_table );
+    pgsql_endCopy(way_table);
 
     char tmp[16];
     snprintf(tmp, sizeof(tmp), "%" PRIdOSMID, id);
     paramValues[0] = tmp;
 
-    PGresult *res = pgsql_execPrepared(sql_conn, "get_way", 1, paramValues, PGRES_TUPLES_OK);
+    auto res = pgsql_execPrepared(sql_conn, "get_way", 1, paramValues,
+                                  PGRES_TUPLES_OK);
 
-    if (PQntuples(res) != 1) {
-        PQclear(res);
+    if (PQntuples(res.get()) != 1) {
         return false;
     }
 
-    pgsql_parse_tags( PQgetvalue(res, 0, 1), tags );
+    {
+        osmium::builder::WayBuilder builder(buffer);
+        builder.set_id(id);
 
-    size_t num_nodes = strtoul(PQgetvalue(res, 0, 2), nullptr, 10);
-    idlist_t list;
-    pgsql_parse_nodes( PQgetvalue(res, 0, 0), list);
-    if (num_nodes != list.size()) {
-        fprintf(stderr, "parse_nodes problem for way %s: expected nodes %zu got %zu\n",
-                tmp, num_nodes, list.size());
-        util::exit_nicely();
+        pgsql_parse_nodes(PQgetvalue(res.get(), 0, 0), buffer, builder);
+        pgsql_parse_tags(PQgetvalue(res.get(), 0, 1), buffer, builder);
     }
-    PQclear(res);
 
-    nodes_get_list(nodes, list);
+    buffer.commit();
+
     return true;
 }
 
-size_t middle_pgsql_t::ways_get_list(const idlist_t &ids, idlist_t &way_ids,
-                                  multitaglist_t &tags, multinodelist_t &nodes) const {
-    if (ids.empty())
-        return 0;
-
+size_t middle_pgsql_t::rel_way_members_get(osmium::Relation const &rel,
+                                           rolelist_t *roles,
+                                           osmium::memory::Buffer &buffer) const
+{
     char tmp[16];
-    std::unique_ptr<char[]> tmp2(new (std::nothrow) char[ids.size() * 16]);
     char const *paramValues[1];
 
-    if (tmp2 == nullptr) return 0; //failed to allocate memory, return */
-
-    // create a list of ids in tmp2 to query the database  */
-    sprintf(tmp2.get(), "{");
-    for(idlist_t::const_iterator it = ids.begin(); it != ids.end(); ++it) {
-        snprintf(tmp, sizeof(tmp), "%" PRIdOSMID ",", *it);
-        strncat(tmp2.get(), tmp, sizeof(char)*(ids.size()*16 - 2));
+    // create a list of ids in tmp2 to query the database
+    std::string tmp2("{");
+    for (auto const &m : rel.members()) {
+        if (m.type() == osmium::item_type::way) {
+            snprintf(tmp, sizeof(tmp), "%" PRIdOSMID ",", m.ref());
+            tmp2.append(tmp);
+        }
     }
-    tmp2[strlen(tmp2.get()) - 1] = '}'; // replace last , with } to complete list of ids*/
+
+    if (tmp2.length() == 1) {
+        return 0; // no ways found
+    }
+    // replace last , with } to complete list of ids
+    tmp2[tmp2.length() - 1] = '}'; 
 
     pgsql_endCopy(way_table);
 
     PGconn *sql_conn = way_table->sql_conn;
 
-    paramValues[0] = tmp2.get();
-    PGresult *res = pgsql_execPrepared(sql_conn, "get_way_list", 1, paramValues, PGRES_TUPLES_OK);
-    int countPG = PQntuples(res);
+    paramValues[0] = tmp2.c_str();
+    auto res = pgsql_execPrepared(sql_conn, "get_way_list", 1, paramValues,
+                                  PGRES_TUPLES_OK);
+    int countPG = PQntuples(res.get());
 
     idlist_t wayidspg;
 
     for (int i = 0; i < countPG; i++) {
-        wayidspg.push_back(strtoosmid(PQgetvalue(res, i, 0), nullptr, 10));
+        wayidspg.push_back(
+            strtoosmid(PQgetvalue(res.get(), i, 0), nullptr, 10));
     }
 
-
     // Match the list of ways coming from postgres in a different order
     //   back to the list of ways given by the caller */
-    for(idlist_t::const_iterator it = ids.begin(); it != ids.end(); ++it) {
+    size_t outres = 0;
+    for (auto const &m : rel.members()) {
+        if (m.type() != osmium::item_type::way) {
+            continue;
+        }
         for (int j = 0; j < countPG; j++) {
-            if (*it == wayidspg[j]) {
-                way_ids.push_back(*it);
-                tags.push_back(taglist_t());
-                pgsql_parse_tags(PQgetvalue(res, j, 2), tags.back());
+            if (m.ref() == wayidspg[j]) {
+                {
+                    osmium::builder::WayBuilder builder(buffer);
+                    builder.set_id(m.ref());
 
-                size_t num_nodes = strtoul(PQgetvalue(res, j, 3), nullptr, 10);
-                idlist_t list;
-                pgsql_parse_nodes( PQgetvalue(res, j, 1), list);
-                if (num_nodes != list.size()) {
-                    fprintf(stderr, "parse_nodes problem for way %s: expected nodes %zu got %zu\n",
-                            tmp, num_nodes, list.size());
-                    util::exit_nicely();
+                    pgsql_parse_nodes(PQgetvalue(res.get(), j, 1), buffer,
+                                      builder);
+                    pgsql_parse_tags(PQgetvalue(res.get(), j, 2), buffer,
+                                     builder);
                 }
 
-                nodes.push_back(nodelist_t());
-                nodes_get_list(nodes.back(), list);
-
+                buffer.commit();
+                if (roles) {
+                    roles->emplace_back(m.role());
+                }
+                outres++;
                 break;
             }
         }
     }
 
-    assert(way_ids.size() <= ids.size());
-
-    PQclear(res);
-
-    return way_ids.size();
+    return outres;
 }
 
 
@@ -634,84 +603,80 @@ void middle_pgsql_t::way_changed(osmid_t osm_id)
 
     //keep track of whatever rels this way intersects
     //TODO: dont need to stop the copy above since we are only reading?
-    PGresult* res = pgsql_execPrepared(rel_table->sql_conn, "mark_rels_by_way", 1, paramValues, PGRES_TUPLES_OK );
-    for(int i = 0; i < PQntuples(res); ++i)
-    {
+    auto res = pgsql_execPrepared(rel_table->sql_conn, "mark_rels_by_way", 1,
+                                  paramValues, PGRES_TUPLES_OK);
+    for (int i = 0; i < PQntuples(res.get()); ++i) {
         char *end;
-        osmid_t marked = strtoosmid(PQgetvalue(res, i, 0), &end, 10);
+        osmid_t marked = strtoosmid(PQgetvalue(res.get(), i, 0), &end, 10);
         rels_pending_tracker->mark(marked);
     }
-    PQclear(res);
 }
 
-void middle_pgsql_t::relations_set(osmid_t id, const memberlist_t &members, const taglist_t &tags)
+void middle_pgsql_t::relations_set(osmium::Relation const &rel)
 {
-    taglist_t member_list;
-    char buf[64];
+    idlist_t parts[3];
 
-    idlist_t all_parts, node_parts, way_parts, rel_parts;
-    all_parts.reserve(members.size());
-    node_parts.reserve(members.size());
-    way_parts.reserve(members.size());
-    rel_parts.reserve(members.size());
-
-    for (memberlist_t::const_iterator it = members.begin(); it != members.end(); ++it) {
-        char type = 0;
-        switch (it->type)
-        {
-            case OSMTYPE_NODE:     node_parts.push_back(it->id); type = 'n'; break;
-            case OSMTYPE_WAY:      way_parts.push_back(it->id); type = 'w'; break;
-            case OSMTYPE_RELATION: rel_parts.push_back(it->id); type = 'r'; break;
-            default:
-                fprintf(stderr, "Internal error: Unknown member type %d\n", it->type);
-                util::exit_nicely();
-        }
-        sprintf( buf, "%c%" PRIdOSMID, type, it->id );
-        member_list.push_back(tag_t(buf, it->role));
+    for (auto const &m : rel.members()) {
+        parts[osmium::item_type_to_nwr_index(m.type())].push_back(m.ref());
     }
 
-    all_parts.insert(all_parts.end(), node_parts.begin(), node_parts.end());
-    all_parts.insert(all_parts.end(), way_parts.begin(), way_parts.end());
-    all_parts.insert(all_parts.end(), rel_parts.begin(), rel_parts.end());
-
-    copy_buffer.reserve(all_parts.size() * 10 + member_list.size() * 24
-                        + tags.size() * 24 + 64);
+    copy_buffer.reserve(rel.members().byte_size() * 2 + rel.tags().byte_size() + 128);
 
     // Params: id, way_off, rel_off, parts, members, tags */
     const char *paramValues[6] = { copy_buffer.c_str(), };
     bool copy = rel_table->copyMode;
     char delim = copy ? '\t' : '\0';
 
-    copy_buffer = std::to_string(id);
+    copy_buffer = std::to_string(rel.id());
     copy_buffer+= delim;
 
     paramValues[1] = paramValues[0] + copy_buffer.size();
-    copy_buffer += std::to_string(node_parts.size());
+    copy_buffer += std::to_string(parts[0].size());
     copy_buffer+= delim;
 
     paramValues[2] = paramValues[0] + copy_buffer.size();
-    copy_buffer += std::to_string(node_parts.size() + way_parts.size());
+    copy_buffer += std::to_string(parts[0].size() + parts[1].size());
     copy_buffer+= delim;
 
     paramValues[3] = paramValues[0] + copy_buffer.size();
-    buffer_store_nodes(all_parts);
+    if (rel.members().empty()) {
+        copy_buffer += "{}";
+    } else {
+        copy_buffer += "{";
+        for (int i = 0; i < 3; ++i) {
+            for (auto it : parts[i]) {
+                copy_buffer += std::to_string(it);
+                copy_buffer += ',';
+            }
+        }
+        copy_buffer[copy_buffer.size() - 1] = '}';
+    }
     copy_buffer+= delim;
 
-    if (member_list.size() == 0) {
+    if (rel.members().empty()) {
         paramValues[4] = nullptr;
         copy_buffer += "\\N";
     } else {
         paramValues[4] = paramValues[0] + copy_buffer.size();
-        buffer_store_tags(member_list, copy);
+        copy_buffer += "{";
+        for (auto const &m : rel.members()) {
+            copy_buffer += '"';
+            copy_buffer += osmium::item_type_to_char(m.type());
+            copy_buffer += std::to_string(m.ref());
+            copy_buffer += "\",\"";
+            buffer_store_string(m.role(), copy);
+            copy_buffer += "\",";
+        }
+        copy_buffer[copy_buffer.size() - 1] = '}';
     }
     copy_buffer+= delim;
 
-    if (tags.size() == 0) {
+    if (rel.tags().empty() && !out_options->extra_attributes) {
         paramValues[5] = nullptr;
         copy_buffer += "\\N";
     } else {
         paramValues[5] = paramValues[0] + copy_buffer.size();
-        buffer_store_tags(tags, copy);
+        buffer_store_tags(rel, out_options->extra_attributes, copy);
     }
 
     if (copy) {
@@ -724,7 +689,7 @@ void middle_pgsql_t::relations_set(osmid_t id, const memberlist_t &members, cons
     }
 }
 
-bool middle_pgsql_t::relations_get(osmid_t id, memberlist_t &members, taglist_t &tags) const
+bool middle_pgsql_t::relations_get(osmid_t id, osmium::memory::Buffer &buffer) const
 {
     char tmp[16];
     char const *paramValues[1];
@@ -732,36 +697,29 @@ bool middle_pgsql_t::relations_get(osmid_t id, memberlist_t &members, taglist_t
     taglist_t member_temp;
 
     // Make sure we're out of copy mode */
-    pgsql_endCopy( rel_table );
+    pgsql_endCopy(rel_table);
 
     snprintf(tmp, sizeof(tmp), "%" PRIdOSMID, id);
     paramValues[0] = tmp;
 
-    PGresult *res = pgsql_execPrepared(sql_conn, "get_rel", 1, paramValues, PGRES_TUPLES_OK);
+    auto res = pgsql_execPrepared(sql_conn, "get_rel", 1, paramValues,
+                                  PGRES_TUPLES_OK);
     // Fields are: members, tags, member_count */
 
-    if (PQntuples(res) != 1) {
-        PQclear(res);
+    if (PQntuples(res.get()) != 1) {
         return false;
     }
 
-    pgsql_parse_tags(PQgetvalue(res, 0, 1), tags);
-    pgsql_parse_tags(PQgetvalue(res, 0, 0), member_temp);
+    {
+        osmium::builder::RelationBuilder builder(buffer);
+        builder.set_id(id);
 
-    if (member_temp.size() != strtoul(PQgetvalue(res, 0, 2), nullptr, 10)) {
-        fprintf(stderr, "Unexpected member_count reading relation %" PRIdOSMID "\n", id);
-        util::exit_nicely();
+        pgsql_parse_members(PQgetvalue(res.get(), 0, 0), buffer, builder);
+        pgsql_parse_tags(PQgetvalue(res.get(), 0, 1), buffer, builder);
     }
 
-    PQclear(res);
+    buffer.commit();
 
-    for (taglist_t::const_iterator it = member_temp.begin(); it != member_temp.end(); ++it) {
-        char tag = it->key[0];
-        OsmType type = (tag == 'n')?OSMTYPE_NODE:(tag == 'w')?OSMTYPE_WAY:(tag == 'r')?OSMTYPE_RELATION:((OsmType)-1);
-        members.push_back(member(type,
-                                 strtoosmid(it->key.c_str()+1, nullptr, 10 ),
-                                 it->value));
-    }
     return true;
 }
 
@@ -779,14 +737,13 @@ void middle_pgsql_t::relations_delete(osmid_t osm_id)
 
     //keep track of whatever ways this relation interesects
     //TODO: dont need to stop the copy above since we are only reading?
-    PGresult* res = pgsql_execPrepared(way_table->sql_conn, "mark_ways_by_rel", 1, paramValues, PGRES_TUPLES_OK );
-    for(int i = 0; i < PQntuples(res); ++i)
-    {
+    auto res = pgsql_execPrepared(way_table->sql_conn, "mark_ways_by_rel", 1,
+                                  paramValues, PGRES_TUPLES_OK);
+    for (int i = 0; i < PQntuples(res.get()); ++i) {
         char *end;
-        osmid_t marked = strtoosmid(PQgetvalue(res, i, 0), &end, 10);
+        osmid_t marked = strtoosmid(PQgetvalue(res.get(), i, 0), &end, 10);
         ways_pending_tracker->mark(marked);
     }
-    PQclear(res);
 }
 
 void middle_pgsql_t::iterate_relations(pending_processor& pf)
@@ -820,14 +777,13 @@ void middle_pgsql_t::relation_changed(osmid_t osm_id)
     //keep track of whatever ways and rels these nodes intersect
     //TODO: dont need to stop the copy above since we are only reading?
     //TODO: can we just mark the id without querying? the where clause seems intersect reltable.parts with the id
-    PGresult* res = pgsql_execPrepared(rel_table->sql_conn, "mark_rels", 1, paramValues, PGRES_TUPLES_OK );
-    for(int i = 0; i < PQntuples(res); ++i)
-    {
+    auto res = pgsql_execPrepared(rel_table->sql_conn, "mark_rels", 1,
+                                  paramValues, PGRES_TUPLES_OK);
+    for (int i = 0; i < PQntuples(res.get()); ++i) {
         char *end;
-        osmid_t marked = strtoosmid(PQgetvalue(res, i, 0), &end, 10);
+        osmid_t marked = strtoosmid(PQgetvalue(res.get(), i, 0), &end, 10);
         rels_pending_tracker->mark(marked);
     }
-    PQclear(res);
 }
 
 idlist_t middle_pgsql_t::relations_using_way(osmid_t way_id) const
@@ -840,14 +796,14 @@ idlist_t middle_pgsql_t::relations_using_way(osmid_t way_id) const
     sprintf(buffer, "%" PRIdOSMID, way_id);
     paramValues[0] = buffer;
 
-    PGresult *result = pgsql_execPrepared(rel_table->sql_conn, "rels_using_way",
-                                          1, paramValues, PGRES_TUPLES_OK );
-    const int ntuples = PQntuples(result);
-    idlist_t rel_ids(ntuples);
+    auto result = pgsql_execPrepared(rel_table->sql_conn, "rels_using_way", 1,
+                                     paramValues, PGRES_TUPLES_OK);
+    const int ntuples = PQntuples(result.get());
+    idlist_t rel_ids;
+    rel_ids.resize((size_t) ntuples);
     for (int i = 0; i < ntuples; ++i) {
-        rel_ids[i] = strtoosmid(PQgetvalue(result, i, 0), nullptr, 10);
+        rel_ids[i] = strtoosmid(PQgetvalue(result.get(), i, 0), nullptr, 10);
     }
-    PQclear(result);
 
     return rel_ids;
 }
@@ -1013,10 +969,14 @@ void middle_pgsql_t::start(const options_t *out_options_)
     // staying set for the second.
     build_indexes = !append && !out_options->droptemp;
 
-    cache.reset(new node_ram_cache( out_options->alloc_chunkwise | ALLOC_LOSSY, out_options->cache, out_options->scale));
-    if (out_options->flat_node_cache_enabled) persistent_cache.reset(new node_persistent_cache(out_options, out_options->append, false, cache));
+    cache.reset(new node_ram_cache(out_options->alloc_chunkwise | ALLOC_LOSSY,
+                                   out_options->cache));
 
-    fprintf(stderr, "Mid: pgsql, scale=%d cache=%d\n", out_options->scale, out_options->cache);
+    if (out_options->flat_node_cache_enabled) {
+        persistent_cache.reset(new node_persistent_cache(out_options, cache));
+    }
+
+    fprintf(stderr, "Mid: pgsql, cache=%d\n", out_options->cache);
 
     // We use a connection per table to enable the use of COPY */
     for (auto& table: tables) {
@@ -1041,7 +1001,7 @@ void middle_pgsql_t::start(const options_t *out_options_)
 
         pgsql_exec(sql_conn, PGRES_COMMAND_OK, "SET client_min_messages = WARNING");
         if (dropcreate) {
-            pgsql_exec(sql_conn, PGRES_COMMAND_OK, "DROP TABLE IF EXISTS %s", table.name);
+            pgsql_exec(sql_conn, PGRES_COMMAND_OK, "DROP TABLE IF EXISTS %s CASCADE", table.name);
         }
 
         if (table.start) {
@@ -1081,9 +1041,6 @@ void middle_pgsql_t::commit(void) {
             table.transactionMode = 0;
         }
     }
-    // Make sure the flat nodes are committed to disk or there will be
-    // surprises later.
-    if (out_options->flat_node_cache_enabled) persistent_cache.reset();
 }
 
 void middle_pgsql_t::pgsql_stop_one(table_desc *table)
@@ -1130,23 +1087,17 @@ void middle_pgsql_t::stop(void)
 }
 
 middle_pgsql_t::middle_pgsql_t()
-    : tables(), num_tables(0), node_table(nullptr), way_table(nullptr), rel_table(nullptr),
-      append(false), mark_pending(true), cache(), persistent_cache(), build_indexes(true)
+: num_tables(0), node_table(nullptr), way_table(nullptr), rel_table(nullptr),
+  append(false), mark_pending(true), build_indexes(true)
 {
+    // clang-format off
     /*table = t_node,*/
     tables.push_back(table_desc(
             /*name*/ "%p_nodes",
            /*start*/ "BEGIN;\n",
-#ifdef FIXED_POINT
-          /*create*/ "CREATE %m TABLE %p_nodes (id " POSTGRES_OSMID_TYPE " PRIMARY KEY {USING INDEX TABLESPACE %i}, lat int4 not null, lon int4 not null, tags text[]) {TABLESPACE %t};\n",
+          /*create*/ "CREATE %m TABLE %p_nodes (id " POSTGRES_OSMID_TYPE " PRIMARY KEY {USING INDEX TABLESPACE %i}, lat int4 not null, lon int4 not null) {TABLESPACE %t};\n",
     /*create_index*/ nullptr,
-         /*prepare*/ "PREPARE insert_node (" POSTGRES_OSMID_TYPE ", int4, int4, text[]) AS INSERT INTO %p_nodes VALUES ($1,$2,$3,$4);\n"
-#else
-          /*create*/ "CREATE %m TABLE %p_nodes (id " POSTGRES_OSMID_TYPE " PRIMARY KEY {USING INDEX TABLESPACE %i}, lat double precision not null, lon double precision not null, tags text[]) {TABLESPACE %t};\n",
-    /*create_index*/ nullptr,
-         /*prepare*/ "PREPARE insert_node (" POSTGRES_OSMID_TYPE ", double precision, double precision, text[]) AS INSERT INTO %p_nodes VALUES ($1,$2,$3,$4);\n"
-#endif
-               "PREPARE get_node (" POSTGRES_OSMID_TYPE ") AS SELECT lat,lon,tags FROM %p_nodes WHERE id = $1 LIMIT 1;\n"
+         /*prepare*/ "PREPARE insert_node (" POSTGRES_OSMID_TYPE ", int4, int4) AS INSERT INTO %p_nodes VALUES ($1,$2,$3);\n"
                "PREPARE get_node_list(" POSTGRES_OSMID_TYPE "[]) AS SELECT id, lat, lon FROM %p_nodes WHERE id = ANY($1::" POSTGRES_OSMID_TYPE "[]);\n"
                "PREPARE delete_node (" POSTGRES_OSMID_TYPE ") AS DELETE FROM %p_nodes WHERE id = $1;\n",
 /*prepare_intarray*/ nullptr,
@@ -1193,6 +1144,7 @@ middle_pgsql_t::middle_pgsql_t()
             /*stop*/  "COMMIT;\n",
    /*array_indexes*/ "CREATE INDEX %p_rels_parts ON %p_rels USING gin (parts) WITH (FASTUPDATE=OFF) {TABLESPACE %i};\n"
                          ));
+    // clang-format on
 
     // set up the rest of the variables from the tables.
     num_tables = tables.size();
@@ -1221,10 +1173,7 @@ std::shared_ptr<const middle_query_t> middle_pgsql_t::get_instance() const {
     //NOTE: this is thread safe for use in pending async processing only because
     //during that process they are only read from
     mid->cache = cache;
-    // The persistent cache on the other hand is not thread-safe for reading,
-    // so we create one per instance.
-    if (out_options->flat_node_cache_enabled)
-        mid->persistent_cache.reset(new node_persistent_cache(out_options, 1, true, cache));
+    mid->persistent_cache = persistent_cache;
 
     // We use a connection per table to enable the use of COPY */
     for(int i=0; i<num_tables; i++) {
