@@ -17,16 +17,13 @@
 /* create a hstore column for all tags */
 #define HSTORE_ALL 2
 
-/* Scale is chosen such that 40,000 * SCALE < 2^32          */
-enum { DEFAULT_SCALE = 100 };
-
 /**
  * Database options, not specific to a table
  */
 class database_options_t {
 public:
     database_options_t();
-    std::string db;
+    boost::optional<std::string> db;
     boost::optional<std::string> username;
     boost::optional<std::string> host;
     boost::optional<std::string> password;
@@ -49,7 +46,6 @@ public:
     virtual ~options_t();
 
     std::string prefix; ///< prefix for table names
-    int scale; ///< scale for converting coordinates to fixed point
     std::shared_ptr<reprojection> projection; ///< SRS of projection
     bool append; ///< Append to existing data
     bool slim; ///< In slim mode
@@ -59,8 +55,9 @@ public:
     boost::optional<std::string> tblsmain_data; ///< Pg Tablespace to store main tables (no default TABLESPACE)
     boost::optional<std::string> tblsslim_data; ///< Pg Tablespace to store slim tables (no default TABLESPACE)
     std::string style; ///< style file to use
-    int expire_tiles_zoom; ///< Zoom level for tile expiry list
-    int expire_tiles_zoom_min; ///< Minimum zoom level for tile expiry list
+    uint32_t expire_tiles_zoom = 0; ///< Zoom level for tile expiry list
+    uint32_t expire_tiles_zoom_min =
+        0;                        ///< Minimum zoom level for tile expiry list
     double expire_tiles_max_bbox; ///< Max bbox size in either dimension to expire full bbox for a polygon
     std::string expire_tiles_filename; ///< File name to output expired tiles list to
     int hstore_mode; ///< add an additional hstore column with objects key/value pairs, and what type of hstore column
@@ -75,7 +72,6 @@ public:
     bool unlogged; ///< use unlogged tables where possible
     bool hstore_match_only; ///< only copy rows that match an explicitly listed key
     bool flat_node_cache_enabled;
-    bool excludepoly;
     bool reproject_area;
     boost::optional<std::string> flat_node_file;
     /**
