@@ -1,8 +1,3 @@
-#include "table.hpp"
-#include "options.hpp"
-#include "util.hpp"
-#include "taginfo.hpp"
-
 #include <exception>
 #include <algorithm>
 #include <cstring>
@@ -10,6 +5,12 @@
 #include <utility>
 #include <time.h>
 
+#include "options.hpp"
+#include "table.hpp"
+#include "taginfo.hpp"
+#include "util.hpp"
+#include "wkb.hpp"
+
 using std::string;
 typedef boost::format fmt;
 
@@ -32,7 +33,6 @@ table_t::table_t(const string& conninfo, const string& name, const string& type,
 
     //we use these a lot, so instead of constantly allocating them we predefine these
     single_fmt = fmt("%1%");
-    point_fmt = fmt("POINT(%.15g %.15g)");
     del_fmt = fmt("DELETE FROM %1% WHERE osm_id = %2%");
 }
 
@@ -40,7 +40,7 @@ table_t::table_t(const table_t& other):
     conninfo(other.conninfo), name(other.name), type(other.type), sql_conn(nullptr), copyMode(false), buffer(), srid(other.srid),
     append(other.append), slim(other.slim), drop_temp(other.drop_temp), hstore_mode(other.hstore_mode), enable_hstore_index(other.enable_hstore_index),
     columns(other.columns), hstore_columns(other.hstore_columns), copystr(other.copystr), table_space(other.table_space),
-    table_space_index(other.table_space_index), single_fmt(other.single_fmt), point_fmt(other.point_fmt), del_fmt(other.del_fmt)
+    table_space_index(other.table_space_index), single_fmt(other.single_fmt), del_fmt(other.del_fmt)
 {
     // if the other table has already started, then we want to execute
     // the same stuff to get into the same state. but if it hasn't, then
@@ -108,7 +108,7 @@ void table_t::start()
     //we are making a new table
     if (!append)
     {
-        pgsql_exec_simple(sql_conn, PGRES_COMMAND_OK, (fmt("DROP TABLE IF EXISTS %1%") % name).str());
+        pgsql_exec_simple(sql_conn, PGRES_COMMAND_OK, (fmt("DROP TABLE IF EXISTS %1% CASCADE") % name).str());
     }
 
     /* These _tmp tables can be left behind if we run out of disk space */
@@ -161,7 +161,9 @@ void table_t::start()
     }//appending
     else {
         //check the columns against those in the existing table
-        std::shared_ptr<PGresult> res = pgsql_exec_simple(sql_conn, PGRES_TUPLES_OK, (fmt("SELECT * FROM %1% LIMIT 0") % name).str());
+        auto res =
+            pgsql_exec_simple(sql_conn, PGRES_TUPLES_OK,
+                              (fmt("SELECT * FROM %1% LIMIT 0") % name).str());
         for (auto const &column :  columns) {
             if (PQfnumber(res.get(), ('"' + column.name + '"').c_str()) < 0) {
 #if 0
@@ -216,7 +218,32 @@ void table_t::stop()
 
         fprintf(stderr, "Sorting data and creating indexes for %s\n", name.c_str());
 
-        pgsql_exec_simple(sql_conn, PGRES_COMMAND_OK, (fmt("CREATE TABLE %1%_tmp %2% AS SELECT * FROM %1% ORDER BY ST_GeoHash(ST_Transform(ST_Envelope(way),4326),10) COLLATE \"C\"") % name % (table_space ? "TABLESPACE " + table_space.get() : "")).str());
+        if (srid == "4326") {
+            /* libosmium assures validity of geometries in 4326, so the WHERE can be skipped.
+               Because we know the geom is already in 4326, no reprojection is needed for GeoHashing */
+            pgsql_exec_simple(
+                sql_conn, PGRES_COMMAND_OK,
+                (fmt("CREATE TABLE %1%_tmp %2% AS\n"
+                     "  SELECT * FROM %1%\n"
+                     "    ORDER BY ST_GeoHash(way,10)\n"
+                     "    COLLATE \"C\"") %
+                 name % (table_space ? "TABLESPACE " + table_space.get() : ""))
+                    .str());
+        } else {
+            /* osm2pgsql's transformation from 4326 to another projection could make a geometry invalid,
+               and these need to be filtered. Also, a transformation is needed for geohashing. */
+            pgsql_exec_simple(
+                sql_conn, PGRES_COMMAND_OK,
+                (fmt("CREATE TABLE %1%_tmp %2% AS\n"
+                     "  SELECT * FROM %1%\n"
+                     "    WHERE ST_IsValid(way)\n"
+                     // clang-format off
+                     "    ORDER BY ST_GeoHash(ST_Transform(ST_Envelope(way),4326),10)\n"
+                     // clang-format on
+                     "    COLLATE \"C\"") %
+                 name % (table_space ? "TABLESPACE " + table_space.get() : ""))
+                    .str());
+        }
         pgsql_exec_simple(sql_conn, PGRES_COMMAND_OK, (fmt("DROP TABLE %1%") % name).str());
         pgsql_exec_simple(sql_conn, PGRES_COMMAND_OK, (fmt("ALTER TABLE %1%_tmp RENAME TO %1%") % name).str());
         fprintf(stderr, "Copying %s to cluster by geometry finished\n", name.c_str());
@@ -233,6 +260,30 @@ void table_t::stop()
             fprintf(stderr, "Creating osm_id index on %s\n", name.c_str());
             pgsql_exec_simple(sql_conn, PGRES_COMMAND_OK, (fmt("CREATE INDEX %1%_pkey ON %1% USING BTREE (osm_id) %2%") % name %
                 (table_space_index ? "TABLESPACE " + table_space_index.get() : "")).str());
+            if (srid != "4326") {
+                pgsql_exec_simple(
+                    sql_conn, PGRES_COMMAND_OK,
+                    (fmt("CREATE OR REPLACE FUNCTION %1%_osm2pgsql_valid()\n"
+                         "RETURNS TRIGGER AS $$\n"
+                         "BEGIN\n"
+                         "  IF ST_IsValid(NEW.way) THEN \n"
+                         "    RETURN NEW;\n"
+                         "  END IF;\n"
+                         "  RETURN NULL;\n"
+                         "END;"
+                         "$$ LANGUAGE plpgsql;") %
+                     name)
+                        .str());
+
+                pgsql_exec_simple(
+                    sql_conn, PGRES_COMMAND_OK,
+                    (fmt("CREATE TRIGGER %1%_osm2pgsql_valid BEFORE INSERT OR UPDATE\n"
+                         "  ON %1%\n"
+                         "    FOR EACH ROW EXECUTE PROCEDURE %1%_osm2pgsql_valid();") %
+                     name)
+                        .str());
+            }
+            //pgsql_exec_simple(sql_conn, PGRES_COMMAND_OK, (fmt("ALTER TABLE %1% ADD CHECK (ST_IsValid(way));") % name).str());
         }
         /* Create hstore index if selected */
         if (enable_hstore_index) {
@@ -247,7 +298,6 @@ void table_t::stop()
             }
         }
         fprintf(stderr, "Creating indexes on %s finished\n", name.c_str());
-        pgsql_exec_simple(sql_conn, PGRES_COMMAND_OK, (fmt("GRANT SELECT ON %1% TO PUBLIC") % name).str());
         pgsql_exec_simple(sql_conn, PGRES_COMMAND_OK, (fmt("ANALYZE %1%") % name).str());
         time(&end);
         fprintf(stderr, "All indexes on %s created in %ds\n", name.c_str(), (int)(end - start));
@@ -259,7 +309,6 @@ void table_t::stop()
 
 void table_t::stop_copy()
 {
-    PGresult* res;
     int stop;
 
     //we werent copying anyway
@@ -278,28 +327,20 @@ void table_t::stop_copy()
        throw std::runtime_error((fmt("stop COPY_END for %1% failed: %2%\n") % name % PQerrorMessage(sql_conn)).str());
 
     //get the result
-    res = PQgetResult(sql_conn);
-    if (PQresultStatus(res) != PGRES_COMMAND_OK)
-    {
-        PQclear(res);
+    pg_result_t res(PQgetResult(sql_conn));
+    if (PQresultStatus(res.get()) != PGRES_COMMAND_OK) {
         throw std::runtime_error((fmt("result COPY_END for %1% failed: %2%\n") % name % PQerrorMessage(sql_conn)).str());
     }
-    PQclear(res);
     copyMode = false;
 }
 
-void table_t::write_node(const osmid_t id, const taglist_t &tags, double lat, double lon)
-{
-    write_row(id, tags, (point_fmt % lon % lat).str());
-}
-
 void table_t::delete_row(const osmid_t id)
 {
     stop_copy();
     pgsql_exec_simple(sql_conn, PGRES_COMMAND_OK, (del_fmt % name % id).str());
 }
 
-void table_t::write_row(const osmid_t id, const taglist_t &tags, const std::string &geom)
+void table_t::write_row(osmid_t id, taglist_t const &tags, std::string const &geom)
 {
     //add the osm id
     buffer.append((single_fmt % id).str());
@@ -312,7 +353,7 @@ void table_t::write_row(const osmid_t id, const taglist_t &tags, const std::stri
         used.assign(tags.size(), false);
 
     //get the regular columns' values
-    write_columns(tags, buffer, hstore_mode == HSTORE_NORM?&used:nullptr);
+    write_columns(tags, buffer, hstore_mode == HSTORE_NORM ? &used : nullptr);
 
     //get the hstore columns' values
     write_hstore_columns(tags, buffer);
@@ -321,25 +362,20 @@ void table_t::write_row(const osmid_t id, const taglist_t &tags, const std::stri
     if (hstore_mode != HSTORE_NONE)
         write_tags_column(tags, buffer, used);
 
-    //give the geometry an srid
-    buffer.append("SRID=");
-    buffer.append(srid);
-    buffer.push_back(';');
-    //add the geometry
-    buffer.append(geom);
+    //add the geometry - encoding it to hex along the way
+    ewkb::writer_t::write_as_hex(buffer, geom);
+
     //we need \n because we are copying from stdin
     buffer.push_back('\n');
 
     //tell the db we are copying if for some reason we arent already
-    if (!copyMode)
-    {
+    if (!copyMode) {
         pgsql_exec_simple(sql_conn, PGRES_COPY_IN, copystr);
         copyMode = true;
     }
 
     //send all the data to postgres
-    if(buffer.length() > BUFFER_SEND_SIZE)
-    {
+    if (buffer.length() > BUFFER_SEND_SIZE) {
         pgsql_CopyData(name.c_str(), sql_conn, buffer);
         buffer.clear();
     }
@@ -526,6 +562,8 @@ table_t::wkb_reader table_t::get_wkb_reader(const osmid_t id)
 
     //the prepared statement get_wkb will behave differently depending on the sql_conn
     //each table has its own sql_connection with the get_way referring to the appropriate table
-    PGresult* res = pgsql_execPrepared(sql_conn, "get_wkb", 1, (const char * const *)paramValues, PGRES_TUPLES_OK);
-    return wkb_reader(res);
+    auto res =
+        pgsql_execPrepared(sql_conn, "get_wkb", 1,
+                           (const char *const *)paramValues, PGRES_TUPLES_OK);
+    return wkb_reader(std::move(res));
 }
